@@ -85,6 +85,7 @@ def train_from_folder(
     generate = False,
     generate_interpolation = False,
     hamburger_res_layers = [32],
+    sle_spatial = False,
     disc_output_size = 1,
     interpolation_num_steps = 100,
     save_frames = False,
@@ -104,6 +105,7 @@ def train_from_folder(
         batch_size = batch_size,
         gradient_accumulate_every = gradient_accumulate_every,
         hamburger_res_layers = cast_list(hamburger_res_layers),
+        sle_spatial = sle_spatial,
         disc_output_size = disc_output_size,
         image_size = image_size,
         optimizer = optimizer,
