@@ -30,7 +30,6 @@ from einops import rearrange
 from pytorch_fid import fid_score
 
 from hamburger_pytorch import Hamburger
-
 from adabelief_pytorch import AdaBelief
 
 # asserts
@@ -174,9 +173,6 @@ def resize_to_minimum_size(min_size, image):
         return torchvision.transforms.functional.resize(image, min_size)
     return image
 
-def normalize_image(img):
-    return img * 2 - 1
-
 class ImageDataset(Dataset):
     def __init__(self, folder, image_size, transparent = False, aug_prob = 0.):
         super().__init__()
@@ -194,8 +190,7 @@ class ImageDataset(Dataset):
             transforms.Resize(image_size),
             RandomApply(aug_prob, transforms.RandomResizedCrop(image_size, scale=(0.5, 1.0), ratio=(0.98, 1.02)), transforms.CenterCrop(image_size)),
             transforms.ToTensor(),
-            transforms.Lambda(expand_greyscale(transparent)),
-            transforms.Lambda(normalize_image)
+            transforms.Lambda(expand_greyscale(transparent))
         ])
 
     def __len__(self):
@@ -232,8 +227,8 @@ class AugWrapper(nn.Module):
 
 norm_class = nn.BatchNorm2d
 
-def upsample():
-    return nn.Upsample(scale_factor = 2)
+def upsample(scale_factor = 2):
+    return nn.Upsample(scale_factor = scale_factor)
 
 # classes
 
@@ -245,15 +240,44 @@ class SLE(nn.Module):
         chan_out
     ):
         super().__init__()
+        self.avg_pool = nn.AdaptiveAvgPool2d((4, 4))
+        self.max_pool = nn.AdaptiveMaxPool2d((4, 4))
+
+        chan_intermediate = chan_in // 2
         self.net = nn.Sequential(
-            nn.AdaptiveAvgPool2d((4, 4)),
-            nn.Conv2d(chan_in, chan_out, 4),
+            nn.Conv2d(chan_in * 2, chan_intermediate, 4),
             nn.LeakyReLU(0.1),
-            nn.Conv2d(chan_out, chan_out, 1),
+            nn.Conv2d(chan_intermediate, chan_out, 1),
             nn.Sigmoid()
         )
     def forward(self, x):
-        return self.net(x)
+        pooled_avg = self.avg_pool(x)
+        pooled_max = self.max_pool(x)
+        return self.net(torch.cat((pooled_max, pooled_avg), dim = 1))
+
+class SpatialSLE(nn.Module):
+    def __init__(self, upsample_times):
+        super().__init__()
+        self.net = nn.Sequential(
+            nn.Conv2d(4, 4, 3, padding = 1),
+            upsample(2 ** upsample_times),
+            nn.Conv2d(4, 4, 3, padding = 1),
+            nn.LeakyReLU(0.1),
+            nn.Conv2d(4, 1, 3, padding = 1),
+            nn.Sigmoid()
+        )
+    def forward(self, x):
+        b, c, h, w = x.shape
+        num_groups = 2
+        mult = math.ceil(c / num_groups)
+        padding = (mult - c % mult) // 2
+        x_padded = F.pad(x, (0, 0, 0, 0, padding, padding))
+        x = rearrange(x, 'b (g c) h w -> b g c h w', g = num_groups)
+
+        pooled_avg = x.mean(dim = 2)
+        pooled_max, _ = x.max(dim = 2)
+        pooled = torch.cat((pooled_avg, pooled_max), dim = 1)
+        return self.net(pooled)
 
 class Generator(nn.Module):
     def __init__(
@@ -264,7 +288,8 @@ class Generator(nn.Module):
         fmap_max = 512,
         fmap_inverse_coef = 12,
         transparent = False,
-        hamburger_res_layers = []
+        hamburger_res_layers = [],
+        sle_spatial = False
     ):
         super().__init__()
         resolution = log2(image_size)
@@ -294,8 +319,10 @@ class Generator(nn.Module):
         self.sle_map = list(filter(lambda t: t[0] <= resolution and t[1] <= resolution, self.sle_map))
         self.sle_map = dict(self.sle_map)
 
-        for (resolution, (chan_in, chan_out)) in zip(self.res_layers, in_out_features):
-            image_width = 2 ** resolution
+        self.num_layers_spatial_res = 4
+
+        for (res, (chan_in, chan_out)) in zip(self.res_layers, in_out_features):
+            image_width = 2 ** res
 
             hamburger = None
             if image_width in hamburger_res_layers:
@@ -305,8 +332,8 @@ class Generator(nn.Module):
                 )
 
             sle = None
-            if resolution in self.sle_map:
-                residual_layer = self.sle_map[resolution]
+            if res in self.sle_map:
+                residual_layer = self.sle_map[res]
                 sle_chan_out = self.res_to_feature_map[residual_layer - 1][-1]
 
                 sle = SLE(
@@ -314,6 +341,10 @@ class Generator(nn.Module):
                     chan_out = sle_chan_out
                 )
 
+            sle_spatial = None
+            if res <= (resolution - self.num_layers_spatial_res):
+                sle_spatial = SpatialSLE(upsample_times = self.num_layers_spatial_res)
+
             layer = nn.ModuleList([
                 nn.Sequential(
                     upsample(),
@@ -322,6 +353,7 @@ class Generator(nn.Module):
                     nn.GLU(dim = 1)
                 ),
                 sle,
+                sle_spatial,
                 hamburger
             ])
             self.layers.append(layer)
@@ -331,10 +363,16 @@ class Generator(nn.Module):
     def forward(self, x):
         x = rearrange(x, 'b c -> b c () ()')
         x = self.initial_conv(x)
+        x = F.normalize(x, dim = -1)
 
         residuals = dict()
+        spatial_residuals = dict()
+
+        for (res, (up, sle, sle_spatial, hamburger)) in zip(self.res_layers, self.layers):
+            if exists(sle_spatial):
+                spatial_res = sle_spatial(x)
+                spatial_residuals[res + self.num_layers_spatial_res] = spatial_res
 
-        for (res, (up, sle, hamburger)) in zip(self.res_layers, self.layers):
             if exists(hamburger):
                 x = hamburger(x) + x
 
@@ -349,7 +387,10 @@ class Generator(nn.Module):
             if next_res in residuals:
                 x = x * residuals[next_res]
 
-        return self.out_conv(x).tanh()
+            if next_res in spatial_residuals:
+                x = x * spatial_residuals[next_res]
+
+        return self.out_conv(x)
 
 class SimpleDecoder(nn.Module):
     def __init__(
@@ -527,6 +568,7 @@ class LightweightGAN(nn.Module):
         transparent = False,
         disc_output_size = 5,
         hamburger_res_layers = [],
+        sle_spatial = False,
         ttur_mult = 1.,
         lr = 2e-4,
         rank = 0,
@@ -542,7 +584,8 @@ class LightweightGAN(nn.Module):
             fmap_max = fmap_max,
             fmap_inverse_coef = fmap_inverse_coef,
             transparent = transparent,
-            hamburger_res_layers = hamburger_res_layers
+            hamburger_res_layers = hamburger_res_layers,
+            sle_spatial = sle_spatial
         )
 
         self.G = Generator(**G_kwargs)
@@ -616,6 +659,7 @@ class Trainer():
         mixed_prob = 0.9,
         gradient_accumulate_every = 1,
         hamburger_res_layers = [],
+        sle_spatial = False,
         disc_output_size = 5,
         lr = 2e-4,
         lr_mlp = 1.,
@@ -658,15 +702,17 @@ class Trainer():
         self.lr = lr
         self.ttur_mult = ttur_mult
         self.batch_size = batch_size
-        self.mixed_prob = mixed_prob
+        self.gradient_accumulate_every = gradient_accumulate_every
 
         self.evaluate_every = evaluate_every
         self.save_every = save_every
         self.steps = 0
 
-        self.gradient_accumulate_every = gradient_accumulate_every
+        self.generator_top_k_gamma = 0.99
+        self.generator_top_k_frac = 0.5
 
         self.hamburger_res_layers = hamburger_res_layers
+        self.sle_spatial = sle_spatial
         self.disc_output_size = disc_output_size
 
         self.d_loss = 0
@@ -721,6 +767,7 @@ class Trainer():
             lr = self.lr,
             latent_dim = self.latent_dim,
             hamburger_res_layers = self.hamburger_res_layers,
+            sle_spatial = self.sle_spatial,
             image_size = self.image_size,
             ttur_mult = self.ttur_mult,
             fmap_max = self.fmap_max,
@@ -747,6 +794,9 @@ class Trainer():
         self.transparent = config['transparent']
         self.hamburger_res_layers = config['hamburger_res_layers']
         self.syncbatchnorm = config['syncbatchnorm']
+        self.disc_output_size = config['disc_output_size']
+        self.sle_spatial = config.pop('sle_spatial', False)
+        self.optimizer = config.pop('optimizer', 'adam')
         self.fmap_max = config.pop('fmap_max', 512)
         del self.GAN
         self.init_GAN()
@@ -756,7 +806,10 @@ class Trainer():
             'image_size': self.image_size,
             'transparent': self.transparent,
             'syncbatchnorm': self.syncbatchnorm,
+            'disc_output_size': self.disc_output_size,
+            'optimizer': self.optimizer,
             'hamburger_res_layers': self.hamburger_res_layers,
+            'sle_spatial': self.sle_spatial
         }
 
     def set_data_src(self, folder):
@@ -838,7 +891,14 @@ class Trainer():
             latents = torch.randn(batch_size, latent_dim).cuda(self.rank)
             generated_images = G(latents)
             fake_output, _ = D_aug(generated_images, **aug_kwargs)
-            fake_output_loss = fake_output
+            fake_output_loss = fake_output.mean(dim = 1)
+
+            epochs = (self.steps * batch_size * self.gradient_accumulate_every) / len(self.dataset)
+            k_frac = max(self.generator_top_k_gamma ** epochs, self.generator_top_k_frac)
+            k = math.ceil(batch_size * k_frac)
+
+            if k != batch_size:
+                fake_output_loss, _ = fake_output_loss.topk(k=k, largest=False)
 
             loss = fake_output_loss.mean()
             gen_loss = loss
@@ -877,7 +937,7 @@ class Trainer():
             if self.steps % self.save_every == 0:
                 self.save(self.checkpoint_num)
 
-            if self.steps % self.evaluate_every == 0 or (self.steps % 100 == 0 and self.steps < 2500):
+            if self.steps % self.evaluate_every == 0 or (self.steps % 100 == 0 and self.steps < 20000):
                 self.evaluate(floor(self.steps / self.evaluate_every))
 
             if exists(self.calculate_fid_every) and self.steps % self.calculate_fid_every == 0 and self.steps != 0:
@@ -954,7 +1014,7 @@ class Trainer():
     @torch.no_grad()
     def generate_truncated(self, G, style, trunc_psi = 0.75, num_image_tiles = 8):
         generated_images = evaluate_in_chunks(self.batch_size, G, style)
-        return (generated_images + 1) * 0.5
+        return generated_images.clamp_(0., 1.)
 
     @torch.no_grad()
     def generate_interpolation(self, num = 0, num_image_tiles = 8, trunc = 1.0, num_steps = 100, save_frames = False):
